==> file4life-1.1.3-source.tgz: file4life-1.1.3/src/F4L.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L where

import F4L.Utils( hashFromS, hashToS )

import Control.DeepSeq
import Control.Monad
import qualified Data.Binary as B
import Data.Binary( Binary )
import Data.Binary.Get()
import qualified Data.ByteString.Short as SBS
import Data.Function( on )
import Data.List( break, unfoldr )
import Data.Word( Word8, Word32 )
import System.FilePath( joinPath, splitDirectories )


storageIdFile = ".f4l-id"


{---------------------------------------- types ----------------------------------------}

-- | Data identifying a file (as ordered sequence of 8-bit bytes)
data FileId = FileId {
    fid_length :: {-# UNPACK #-} !Int
  , fid_hash :: {-# UNPACK #-} !SBS.ShortByteString 
  } deriving( Eq, Ord )

type EnvironmentId = String

type StorageId = String

type SessionId = String

data SessionVersion = V0 | V0_1
                    deriving ( Eq, Show )


currentVersion = V0_1


-- | Space (ex. partition or dir) dedicated to storage of files
data Storage = Storage { st_id :: !StorageId, st_path :: !FilePath }

-- | Database for metadata
data MetaDB = MetaDB { mdb_file :: !FilePath }
            deriving Show

type TimeStamp = Int


data Priority = Skip | DefaultPriority | Priority !Int | AsManyAsPossible
              deriving (Eq, Show)

instance Ord Priority where
  compare = compare `on` toN
    where
      toN Skip = -1
      toN DefaultPriority = 1
      toN ( Priority i ) = i
      toN AsManyAsPossible = 10


-- | Attributes attached to a file or path
data FilePathAttr a =
    Require !Priority
  | Tag !String -- ^ a tag
  | Property !String !String
  deriving( Eq, Show )

data FPFile

instance Eq FPFile

instance Show FPFile

data FPPath

instance Eq FPPath -- Needed for automatic instance in FilePathAttr

instance Show FPPath -- Needed for automatic instance in FilePathAttr

type FileAttr = FilePathAttr FPFile -- Needed for automatic instance in FilePathAttr

type PathAttr = FilePathAttr FPPath -- Needed for automatic instance in FilePathAttr

data SessionData = SessionData { 
    sdat_ver :: SessionVersion 
  , sdat_id :: SessionId 
  , sdat_env :: EnvironmentId 
  , sdat_ts :: TimeStamp 
  , sdat_len :: Int } deriving Show


formatSession ( SessionData a b c d e ) =
  unwords [ show a, b, c, show d, show e ]


{---------------------------------------- Show ----------------------------------------}
  
instance Show FileId where
  show( FileId l h ) = "FileId(" ++ show l ++ ", " ++ show ( hashToS h ) ++ ")"

-- | This considers files saved in 4k blocks
fidDiskSpace :: FileId -> Int
fidDiskSpace ( FileId l _ ) = ( (l-1) `div` 4096 + 1 ) * 4096


-- | return ( destination dir, destination filename )
fidToBDest :: FileId -> ( FilePath, FilePath )
fidToBDest fid =
  let h = hashToS $ fid_hash fid
  in ( joinPath [ take 2 h, take 2 $ drop 2 h ], h ++ "_" ++ show ( fid_length fid ) )


-- | path argument is relative to backup dir
fidFromBDest :: FilePath -> Maybe FileId
fidFromBDest path = do
  (d1, d2, filename) <- case splitDirectories path of
    [d1, d2, filename] -> Just (d1, d2, filename)
    _ -> Nothing
  (h, post) <- case break ( (==) '_' ) filename of
    (h, _ : post) -> Just (h, post)
    _ -> Nothing
  l <- case reads post of
    [(l, "")] -> Just l
    _ -> Nothing
  h <- hashFromS h
  let fid = FileId l h
  let (p, fn) = fidToBDest fid
  if joinPath [p, fn] == path then Just fid else Nothing


testFid :: Int -> FileId
testFid i =
  FileId i $ SBS.pack $ take 20 $ map fromIntegral $ unfoldr ( Just . swap . ( `divMod` 256 ) ) i
  where
    swap (a, b) = (b, a)


{---------------------------------------- NFData ----------------------------------------}

instance NFData FileId

instance NFData Priority

instance NFData ( FilePathAttr a ) where
  rnf ( Tag s ) = rnf s
  rnf ( Property name value ) = rnf name `seq` rnf value
  rnf x = x `seq` ()


{---------------------------------------- Conversion to/from binary ----------------------------------------}

instance Binary SBS.ShortByteString where
  put = B.put . SBS.unpack
  get = liftM SBS.pack B.get

instance Binary FileId where
  put( FileId l h ) = B.put ( fromIntegral l :: Word32 ) >> mapM_ B.putWord8 ( SBS.unpack h )
  get = liftM2 FileId ( liftM fromW32 B.get ) ( liftM SBS.pack $ replicateM 20 B.get )
    where 
      fromW32 :: Word32 -> Int 
      fromW32 = fromIntegral

instance Binary Priority where
  put Skip = B.put '0'
  put DefaultPriority = B.put 'a'
  put ( Priority i ) = B.put 'p' >> B.put i -- ( toEnum ( i + fromEnum '0' ) :: Char )
  put AsManyAsPossible = B.put 'z'
  get = do
    t <- B.get
    case t of
      '0' -> return Skip
      'a' -> return DefaultPriority
      'p' -> liftM Priority B.get
      'z' -> return AsManyAsPossible
        
instance Binary ( FilePathAttr a ) where
  put( Require req ) = B.put 'r' >> B.put req
  put( Tag s ) = B.put 't' >> B.put s
  put( Property k v ) = B.put 'p' >> B.put k >> B.put v
  get = do
    t <- B.get
    case t of
      'r' -> liftM Require B.get
      't' -> liftM Tag B.get
      'p' -> liftM2 Property B.get B.get

codeV0 :: Word8 
codeV0 = toEnum $ fromEnum '0'

codeV0_1 :: Word8 
codeV0_1 = toEnum $ fromEnum '0' + 10

instance Binary SessionVersion where
  put V0 = B.put codeV0
  put V0_1 = B.put codeV0_1
  get = B.get >>= \ c ->
    case c of
      x | x == codeV0 -> return V0
      x | x == codeV0_1 -> return V0_1
      x -> fail $ "Error unserializing SessionVersion: unknown version: " ++ show ( fromEnum x )

==> file4life-1.1.4-source.tgz: file4life-1.1.4/src/F4L.hs <==
{- This file is part of file4life, a file backup program written in Haskell.
 - file4life is Copyright (C) Antonio Bellezza 2014
 - file4life is free software: you can redistribute it and/or modify
 - it under the terms of the GNU General Public License as published by
 - the Free Software Foundation, either version 3 of the License, or
 - (at your option) any later version.

 - file4life is distributed in the hope that it will be useful,
 - but WITHOUT ANY WARRANTY; without even the implied warranty of
 - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 - GNU General Public License for more details.

 - You should have received a copy of the GNU General Public License
 - along with file4life.  If not, see <http://www.gnu.org/licenses/>.
 -}
module F4L where

import F4L.Utils( hashFromS, hashToS )

import Control.DeepSeq
import Control.Monad
import qualified Data.Binary as B
import Data.Binary( Binary )
import Data.Binary.Get()
import qualified Data.ByteString.Short as SBS
import Data.Function( on )
import Data.List( break, unfoldr )
import Data.Word( Word8, Word32 )
import System.FilePath( joinPath, splitDirectories )


storageIdFile = ".f4l-id"


{---------------------------------------- types ----------------------------------------}

-- | Data identifying a file (as ordered sequence of 8-bit bytes)
data FileId = FileId {
    fid_length :: {-# UNPACK #-} !Int
  , fid_hash :: {-# UNPACK #-} !SBS.ShortByteString 
  } deriving( Eq, Ord )

type EnvironmentId = String

type StorageId = String

type SessionId = String

-- | Each change in the Op formats/encodings should have a new version identifier
data SessionVersion = V0 | V0_1
                    deriving ( Eq, Show )


currentVersion = V0_1


-- | Database for metadata
data MetaDB = MetaDB { mdb_file :: !FilePath }
            deriving Show

-- | unix timestamp, rounded to second
type TimeStamp = Int


data Priority = Skip | DefaultPriority | Priority !Int | AsManyAsPossible
              deriving (Eq, Show)

instance Ord Priority where
  compare = compare `on` toN
    where
      toN Skip = -1
      toN DefaultPriority = 1
      toN ( Priority i ) = i
      toN AsManyAsPossible = 10


-- | Attributes attached to a file or path
data FilePathAttr a =
    Require !Priority
  | Tag !String -- ^ a tag
  | Property !String !String
  deriving( Eq, Show )

data FPFile

instance Eq FPFile

instance Show FPFile

data FPPath

instance Eq FPPath -- Needed for automatic instance in FilePathAttr

instance Show FPPath -- Needed for automatic instance in FilePathAttr

type FileAttr = FilePathAttr FPFile -- Needed for automatic instance in FilePathAttr

type PathAttr = FilePathAttr FPPath -- Needed for automatic instance in FilePathAttr

data SessionData = SessionData { 
    sdat_ver :: SessionVersion 
  , sdat_id :: SessionId 
  , sdat_env :: EnvironmentId 
  , sdat_ts :: TimeStamp 
  , sdat_len :: Int } deriving Show


formatSession ( SessionData a b c d e ) =
  unwords [ show a, b, c, show d, show e ]


{---------------------------------------- Show ----------------------------------------}
  
instance Show FileId where
  show( FileId l h ) = "FileId(" ++ show l ++ ", " ++ show ( hashToS h ) ++ ")"


-- | This considers files saved in 4k blocks
fidDiskSpace :: FileId -> Int
fidDiskSpace ( FileId l _ ) = ( (l-1) `div` 4096 + 1 ) * 4096


-- | return ( destination dir, destination filename ) for fid within a backup directory
fidToBDest :: FileId -> ( FilePath, FilePath )
fidToBDest fid =
  let h = hashToS $ fid_hash fid
  in ( joinPath [ take 2 h, take 2 $ drop 2 h ], h ++ "_" ++ show ( fid_length fid ) )


-- | path argument is relative to backup dir
fidFromBDest :: FilePath -> Maybe FileId
fidFromBDest path = do
  (d1, d2, filename) <- case splitDirectories path of
    [d1, d2, filename] -> Just (d1, d2, filename)
    _ -> Nothing
  (h, post) <- case break ( (==) '_' ) filename of
    (h, _ : post) -> Just (h, post)
    _ -> Nothing
  l <- case reads post of
    [(l, "")] -> Just l
    _ -> Nothing
  h <- hashFromS h
  let fid = FileId l h
  let (p, fn) = fidToBDest fid
  if joinPath [p, fn] == path then Just fid else Nothing


testFid :: Int -> FileId
testFid i =
  FileId i $ SBS.pack $ take 20 $ map fromIntegral $ unfoldr ( Just . swap . ( `divMod` 256 ) ) i
  where
    swap (a, b) = (b, a)


{---------------------------------------- NFData ----------------------------------------}

instance NFData FileId

instance NFData Priority

instance NFData ( FilePathAttr a ) where
  rnf ( Tag s ) = rnf s
  rnf ( Property name value ) = rnf name `seq` rnf value
  rnf x = x `seq` ()


{---------------------------------------- Conversion to/from binary ----------------------------------------}

instance Binary SBS.ShortByteString where
  put = B.put . SBS.unpack
  get = liftM SBS.pack B.get

instance Binary FileId where
  put( FileId l h ) = B.put ( fromIntegral l :: Word32 ) >> mapM_ B.putWord8 ( SBS.unpack h )
  get = liftM2 FileId ( liftM fromW32 B.get ) ( liftM SBS.pack $ replicateM 20 B.get )
    where 
      fromW32 :: Word32 -> Int 
      fromW32 = fromIntegral

instance Binary Priority where
  put Skip = B.put '0'
  put DefaultPriority = B.put 'a'
  put ( Priority i ) = B.put 'p' >> B.put i -- ( toEnum ( i + fromEnum '0' ) :: Char )
  put AsManyAsPossible = B.put 'z'
  get = do
    t <- B.get
    case t of
      '0' -> return Skip
      'a' -> return DefaultPriority
      'p' -> liftM Priority B.get
      'z' -> return AsManyAsPossible
        
instance Binary ( FilePathAttr a ) where
  put( Require req ) = B.put 'r' >> B.put req
  put( Tag s ) = B.put 't' >> B.put s
  put( Property k v ) = B.put 'p' >> B.put k >> B.put v
  get = do
    t <- B.get
    case t of
      'r' -> liftM Require B.get
      't' -> liftM Tag B.get
      'p' -> liftM2 Property B.get B.get

codeV0 :: Word8 
codeV0 = toEnum $ fromEnum '0'

codeV0_1 :: Word8 
codeV0_1 = toEnum $ fromEnum '0' + 10

instance Binary SessionVersion where
  put V0 = B.put codeV0
  put V0_1 = B.put codeV0_1
  get = B.get >>= \ c ->
    case c of
      x | x == codeV0 -> return V0
      x | x == codeV0_1 -> return V0_1
      x -> fail $ "Error unserializing SessionVersion: unknown version: " ++ show ( fromEnum x )
